import math
from typing import List, Tuple


def sub_vector(v1: List[float], v2: List[float]) -> List:
    return [v1[0] - v2[0], v1[1] - v2[1], v1[2] - v2[2]]


def cross_product(v1: List[float], v2: List[float]) -> List[float]:
    return [
        v1[1] * v2[2] - v1[2] * v2[1],
        v1[2] * v2[0] - v1[0] * v2[2],
        v1[0] * v2[1] - v1[1] * v2[0],
    ]


def vector_norm(v: List[float]) -> float:
    return math.sqrt((v[0] * v[0]) + (v[1] * v[1]) + (v[2] * v[2]))


def normalize_vector(v: List[float]) -> List[float]:
    vn = vector_norm(v)
    if vn == 0.0:
        return v
    invlen = 1.0 / vn
    res = [n * invlen for n in v]

    return res


def plane_normal(v1: List[float], v2: List[float], v3: List[float]) -> List[float]:
    return normalize_vector(cross_product(sub_vector(v2, v1), sub_vector(v3, v1)))


def vector_transform(v: List[float], matrix: List[List[float]]) -> List[float]:
    mx = matrix[0]
    my = matrix[1]
    mz = matrix[2]
    mw = matrix[3]

    rx = v[0] * mx[0] + v[1] * my[0] + v[2] * mz[0]
    ry = v[0] * mx[1] + v[1] * my[1] + v[2] * mz[1]
    rz = v[0] * mx[2] + v[1] * my[2] + v[2] * mz[2]

    if len(v) == 3:
        rx += mw[0]
        ry += mw[1]
        rz += mw[2]
        return [rx, ry, rz]

    rx += v[3] * mw[0]
    ry += v[3] * mw[1]
    rz += v[3] * mw[2]
    rw = v[0] * mx[3] + v[1] * my[3] + v[2] * mz[3] + v[3] * mw[3]
    return [rx, ry, rz, rw]


def invert_vector(v: List[float]) -> List[float]:
    return [-v[0], -v[1], -v[2]]


def min_max(vlist: List[List[float]]) -> Tuple[List[float], List[float]]:
    x_list = [item[0] for item in vlist]
    y_list = [item[1] for item in vlist]
    z_list = [item[2] for item in vlist]
    return (
        [min(x_list), min(y_list), min(z_list)],
        [max(x_list), max(y_list), max(z_list)],
    )


def scale_vector(v: List[float], d: float) -> List[float]:
    return [q * d for q in v]


def add_vectors(v1: List[float], v2: List[float]) -> List[float]:
    if len(v1) != len(v2):
        raise BaseException("Vector lengths must be equal")
    result = []
    for i in range(len(v1)):
        result.append(v1[i] + v2[i])
    return result


def dot_product(v1: List[float], v2: List[float]) -> float:
    """Return vector dot product of v1 and v2"""
    return v1[0] * v2[0] + v1[1] * v2[1] + v1[2] * v2[2]


def vector_angle(v1: List[float], v2: List[float]) -> float:
    v1_n = normalize_vector(v1)
    v2_n = normalize_vector(v2)
    return math.acos(dot_product(v1_n, v2_n) / (vector_norm(v1_n) * vector_norm(v2_n)))


def _ensure_affine(v: List[float]):
    res = v.copy()
    if len(res) == 2:
        res.append(0)
    return res


def mid_point(v1: List[float], v2: List[float]) -> List[float]:
    v1_c = _ensure_affine(v1)
    v2_c = _ensure_affine(v2)

    diff = sub_vector(v2_c, v1_c)
    leng = vector_norm(diff)
    diff = normalize_vector(diff)
    diff = scale_vector(diff, leng / 2.0)
    result = add_vectors(diff, v1_c)
    if len(v1) == 2:
        return result[:2]
    return result


def distance(v1: List[float], v2: List[float]) -> float:
    v1_c = _ensure_affine(v1)
    v2_c = _ensure_affine(v2)
    diff = sub_vector(v2_c, v1_c)
    return vector_norm(diff)


def rotate_around_z(v: List[float], angle: float) -> List[float]:
    s = math.sin(angle)
    c = math.cos(angle)
    return [v[0] * c + v[1] * s, v[1] * c - v[0] * s, v[2]]



import logging
from itertools import combinations
from typing import Any, Callable, List, Optional, Set, Type

from payton.math.geometry import distance
from payton.scene.geometry.base import Object
from payton.scene.geometry.mesh import Mesh
from payton.scene.geometry.sphere import Sphere


class Collision(object):
    def __init__(self, obj1: Object, obj2: Object) -> None:
        self.object_1 = obj1
        self.object_2 = obj2
        self.segment = 1


class CollisionTest(object):
    """
    TODO: Improve collision detection algorithms.
    """

    SPHERICAL = 0
    AABB = 1

    def __init__(
        self, callback: Callable, level: int = AABB, objects: Optional[List[Type[Mesh]]] = None, **kwargs: Any,
    ) -> None:
        self.objects: List[Type[Mesh]] = [] if objects is None else objects
        self.callback: Callable = callback
        self.level: int = level
        self._pairs: List[Set[Mesh]] = []

    def add_object(self, obj: Type[Mesh]) -> None:
        if not isinstance(obj, Mesh):
            logging.error("object must be an instance of Mesh")
            return
        self.objects.append(obj)

    def _dist(self, obj1: Object, obj2: Object) -> float:
        p1 = obj1._model_matrix[3]
        p2 = obj2._model_matrix[3]
        return distance(p1, p2)

    def _bounding_sphere_collision(self, obj1: Object, obj2: Object) -> bool:
        dist = self._dist(obj1, obj2)
        total_radius = obj1.bounding_radius + obj2.bounding_radius
        return dist <= total_radius

    def _sphere_in_sphere_collision(self, obj1: Mesh, obj2: Mesh) -> bool:
        dist = self._dist(obj1, obj2)
        if obj1.bounding_radius > dist + obj2.bounding_radius:
            return True
        if obj2.bounding_radius > dist + obj1.bounding_radius:
            return True
        return False

    def _aabb_collision_test(self, obj1: Mesh, obj2: Mesh) -> bool:
        a_min = obj1.bounding_box[0]
        a_max = obj1.bounding_box[1]
        b_min = obj2.bounding_box[0]
        b_max = obj2.bounding_box[1]
        if a_min[0] >= b_max[0] or b_min[0] >= a_max[0]:
            return False
        if a_min[1] >= b_max[1] or b_min[1] >= a_max[1]:
            return False
        if a_min[2] >= b_max[2] or b_min[2] >= a_max[2]:
            return False
        return True

    def _test(self, obj1: Mesh, obj2: Mesh) -> bool:
        bs_test = self._bounding_sphere_collision(obj1, obj2)
        if not bs_test:
            return False
        if self.level == self.SPHERICAL:
            return True
        if isinstance(obj1, Sphere) and isinstance(obj2, Sphere):
            # If both objects are Spheres, this check is enough
            return True
        if isinstance(obj1, Mesh) and isinstance(obj2, Mesh):
            return self._aabb_collision_test(obj1, obj2)
        return False

    def resolve(self, obj1: Mesh, obj2: Mesh) -> None:
        pair = set([obj1, obj2])
        if pair in self._pairs:
            self._pairs.remove(pair)

    def check(self):
        for pair_tuple in combinations(self.objects, 2):
            pair = set(pair_tuple)
            if pair not in self._pairs:
                res = self._test(pair_tuple[0], pair_tuple[1])
                if res:
                    self._pairs.append(pair)

        if len(self._pairs) > 0:
            self.callback(self, [list(pair) for pair in self._pairs])



# pylama:ignore=C901
import logging
from typing import Any, Optional

import sdl2

from payton.scene.gui import EditBox
from payton.scene.observer import BUTTON_LEFT, BUTTON_MIDDLE, BUTTON_RIGHT


class Controller(object):
    def __init__(self):
        super().__init__()
        self._active_object: Optional[EditBox] = None

    def keyboard(self, event: sdl2.SDL_Event, scene: Any) -> None:
        if event.type == sdl2.SDL_QUIT:
            logging.debug("Quit SDL Scene")
            scene.running = False
            for clock in scene.clocks:
                c = scene.clocks[clock]
                logging.debug(f"Kill clock [{clock}]")
                c.kill()
                c.join()

        if event.type == sdl2.SDL_KEYDOWN:
            key = event.key.keysym.sym
            if key == sdl2.SDLK_LSHIFT:
                scene._shift_down = True
            if key == sdl2.SDLK_LCTRL:
                scene._ctrl_down = True

        if event.type == sdl2.SDL_TEXTINPUT:
            if self._active_object is not None:
                self._active_object._on_keypress(event.text.text.decode('utf-8'))

        if event.type == sdl2.SDL_KEYUP:
            key = event.key.keysym.sym

            # GUI Related Controls, move to another place when you have time
            if self._active_object is not None:
                if key == sdl2.SDLK_ESCAPE or (key == sdl2.SDLK_RETURN and not self._active_object.multiline):
                    self._active_object._exit()
                    self._active_object = None
                    sdl2.SDL_ShowCursor(True)
                    sdl2.SDL_StopTextInput()
                if key == sdl2.SDLK_BACKSPACE and self._active_object is not None:
                    self._active_object.backspace()
                if key == sdl2.SDLK_RETURN and self._active_object is not None and self._active_object.multiline:
                    self._active_object._on_keypress("\n")
                if self._active_object and key == sdl2.SDLK_LEFT:
                    self._active_object.cursor_left()
                if self._active_object and key == sdl2.SDLK_RIGHT:
                    self._active_object.cursor_right()
                return

            if key == sdl2.SDLK_LSHIFT:
                scene._shift_down = False
            if key == sdl2.SDLK_LCTRL:
                scene._ctrl_down = False

            if key == sdl2.SDLK_c:
                # below variable assignment is only for code style
                p = scene.active_observer.perspective
                scene.active_observer.perspective = not p
                logging.debug(f"Observer Perspective={scene.active_observer.perspective}")

            if key == sdl2.SDLK_g:
                scene.grid.visible = not scene.grid.visible

            if key == sdl2.SDLK_SPACE:
                for clock in scene.clocks:
                    c = scene.clocks[clock]
                    logging.debug(f"Pause clock [{clock}]")
                    c.pause()

            if key == sdl2.SDLK_h:
                scene.huds["_help"]._visible = not scene.huds["_help"]._visible

            if key == sdl2.SDLK_w:
                for obj in scene.objects:
                    scene.objects[obj].toggle_wireframe()

            if key == sdl2.SDLK_ESCAPE:
                scene.running = False

            if key in [sdl2.SDLK_F2, sdl2.SDLK_F3]:
                active = 0
                for i in range(len(scene.observers)):
                    if scene.observers[i].active:
                        active = i
                if key == sdl2.SDLK_F2:
                    active -= 1
                if key == sdl2.SDLK_F3:
                    active += 1
                if active < 0:
                    active = len(scene.observers) - 1
                if active > len(scene.observers) - 1:
                    active = 0
                scene.active_observer = scene.observers[active]
                for i in range(len(scene.observers)):
                    scene.observers[i].active = i == active

    def mouse(self, event: sdl2.SDL_Event, scene: Any) -> None:
        observer = scene.active_observer
        if event.type == sdl2.SDL_MOUSEBUTTONUP:
            observer._prev_intersection = None

        if event.type == sdl2.SDL_MOUSEBUTTONDOWN:
            mx, my = event.button.x, event.button.y
            for hud in scene.huds:
                h = scene.huds[hud]
                for shape in h.children:
                    focus_element = h.children[shape].click(mx, my)
                    if focus_element:
                        if self._active_object is not None and self._active_object != focus_element:
                            self._active_object._exit()
                            self._active_object = None
                            sdl2.SDL_ShowCursor(True)
                            sdl2.SDL_StopTextInput()
                        if hasattr(focus_element, "_on_keypress"):
                            self._active_object = focus_element
                            sdl2.SDL_ShowCursor(False)
                            sdl2.SDL_StartTextInput()
                        return

            eye, ray_dir = observer.screen_to_world(mx, my, scene.window_width, scene.window_height)

            list = []
            if callable(scene.on_select):
                for obj in scene.objects:
                    hit = scene.objects[obj].select(eye, ray_dir)
                    if hit:
                        list.append(scene.objects[obj])
                if len(list) > 0:
                    scene.on_select(list)

            if not (scene._shift_down or scene._ctrl_down):
                scene._check_click_plane(eye, ray_dir)

        if event.type == sdl2.SDL_MOUSEWHEEL:
            yrel = event.wheel.y
            observer.mouse_wheel(yrel)

        if event.type == sdl2.SDL_MOUSEMOTION:
            button = -1
            if event.motion.state == sdl2.SDL_BUTTON_LMASK:
                button = BUTTON_LEFT
            if event.motion.state == sdl2.SDL_BUTTON_RMASK:
                button = BUTTON_RIGHT
            if event.motion.state == sdl2.SDL_BUTTON_MIDDLE:
                button = BUTTON_MIDDLE
            observer.mouse_move(
                button,
                scene._shift_down,
                scene._ctrl_down,
                event.motion.x,
                event.motion.y,
                event.motion.xrel,
                event.motion.yrel,
                scene.window_width,
                scene.window_height,
            )



from enum import Enum
from typing import Any, Callable, Dict, Optional, Tuple

import numpy as np

from payton.scene.gui.base import Shape2D, Text
from payton.scene.shader import Shader


class Theme:
    def __init__(self):
        self.opacity = 0.9

        self.background_color: Tuple[float, float, float] = [0.05, 0.05, 0.05]
        self.text_color: Tuple[float, float, float] = [1.0, 1.0, 1.0]
        self.title_background_color: Tuple[float, float, float] = [
            235 / 255,
            210 / 255,
            52 / 255,
        ]
        self.title_text_color: Tuple[float, float, float] = [0.0, 0.0, 0.0]

    def secondary(self):
        self.background_color, self.text_color = self.title_background_color, self.title_text_color
        self.title_background_color, self.title_text_color = self.background_color, self.text_color


class WindowAlignment(Enum):
    FREE = "free"
    LEFT = "left"
    TOP = "top"
    RIGHT = "right"
    BOTTOM = "bottom"


class WindowElement(Shape2D):
    def __init__(
        self,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        **kwargs,
    ):
        kwargs["position"] = (left, top, 0)
        kwargs["size"] = (width, height)
        super().__init__(**kwargs)
        self._init: bool = False
        self.align = align
        self.theme = theme if theme is not None else Theme()
        self.material.opacity = self.theme.opacity

    def _reposition(self):
        if self.align == WindowAlignment.LEFT:
            self.position = [0.0, 0.0, 0.0]
            self.size = (self.size[0], self._parent_height)
        elif self.align == WindowAlignment.RIGHT:
            self.position = [self._parent_width - self.size[0], 0.0, 0.0]
            self.size = (self.size[0], self._parent_height)
        elif self.align == WindowAlignment.TOP:
            self.position = [0.0, 0.0, 0.0]
            self.size = (self._parent_width, self.size[1])
        elif self.align == WindowAlignment.BOTTOM:
            pos = [0, self._parent_height - self.size[1], 0]
            self.position = pos
            self.size = (self._parent_width, self.size[1])

    def draw(self):
        self._reposition()

    def render(
        self, lit: bool, shader: Shader, parent_matrix: Optional[np.ndarray] = None, _primitive: int = None,
    ) -> None:
        if not self._init:
            self.draw()

        super().render(lit, shader, parent_matrix)

    def add_child(self, name, obj: Shape2D) -> bool:  # type: ignore
        if obj.position[0] > self.size[0]:
            obj.position[0] = self.size[0] - 1
        total_v = obj.position[0] + obj.size[0]
        exceed = int(total_v - self.size[0])
        if exceed > 0:
            obj.size = (obj.size[0] - exceed, obj.size[1])

        if obj.position[1] > self.size[1]:
            obj.position[1] = self.size[1] - 1
        total_v = obj.position[1] + obj.size[1]
        exceed = int(total_v - self.size[1])
        if exceed > 0:
            obj.size = (obj.size[0], obj.size[1] - exceed)

        super().add_child(name, obj)
        obj.set_parent_size(self.size[0], self.size[1])


class Window(WindowElement):
    def __init__(
        self,
        title: str = "",
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        **kwargs: Dict[str, Any],
    ):
        """Initialize Window

        @NOTE: I am aware that I could have just define "title" as other
               elemenets are already arguments of WindowElement but some
               code hinting and completion tools are a bit stupid. They
               can not inherit from parent class.
        """
        super().__init__(
            width=width, height=height, left=left, top=top, align=align, theme=theme, **kwargs,
        )
        self.title = title
        self.add_child(
            "title",
            Text(
                position=(0, 0, 0),
                size=(self.size[0], 20),
                label=self.title,
                # bgcolor=self.theme.title_background_color,
                color=self.theme.title_text_color,
                opacity=0.5,
            ),
        )

    def draw(self):
        super().draw()
        w, h = self.size[0], self.size[1]
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 22, 1], [w, h, 1], [w, 22, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )
            self.add_triangle(
                [[0, 22, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )
            self.add_triangle(
                [[0, 0, 1], [w - 1, 22, 1], [w - 1, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[0, 0, 1], [0, 22, 1], [w - 1, 22, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self._init = True


class Panel(WindowElement):
    def draw(self):
        super().draw()
        w, h = self.size[0], self.size[1]
        self.clear_triangles()
        if not self._init:
            self.add_triangle(
                [[0, 0, 1], [w, h, 1], [w, 0, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[0, 0, 1], [0, h, 1], [w, h, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                    self.theme.title_background_color,
                ],
            )
            self.add_triangle(
                [[1, 1, 1], [w - 1, h - 1, 1], [w - 1, 1, 1]],
                texcoords=[[0, 0], [1, 1], [1, 0]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )
            self.add_triangle(
                [[1, 1, 1], [1, h - 1, 1], [w - 1, h - 1, 1]],
                texcoords=[[0, 0], [0, 1], [1, 1]],
                colors=[self.theme.background_color, self.theme.background_color, self.theme.background_color],
            )

            self._init = True


class Button(Panel):
    def __init__(
        self,
        label: str,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        on_click: Optional[Callable] = None,
        **kwargs: Any,
    ):
        kwargs["on_click"] = on_click
        super().__init__(
            width=width, height=height, left=left, top=top, align=align, theme=theme, **kwargs,
        )
        if height < 30:
            height = 30

        self._label = label
        self.text = Text(position=(0, 0, 1), size=(10, 10), label=label, color=self.theme.text_color,)
        self.add_child("label", self.text)

    def draw(self, **kwargs):
        super().draw()
        text_size = self.text.text_size
        x = (self.size[0] - text_size[0]) / 2
        y = (self.size[1] - text_size[1]) / 2
        if x < 0:
            x = 0
        if y < 0:
            y = 0
        y -= 4
        self.text.label = self._label
        self.text.position = [x, y]
        crop = [0, 0, text_size[0], text_size[1] + 4]
        if text_size[0] > self.size[0]:
            crop[0] = (text_size[0] - self.size[0]) / 2.0
            crop[2] = text_size[0] - crop[0]
            diff = int(crop[2] - crop[0])
            text_size = (diff, text_size[1])
        if text_size[1] > self.size[1]:
            crop[1] = (text_size[1] - self.size[1]) / 2.0
            crop[3] = text_size[1] - crop[1]
            text_size = (text_size[0], self.size[1])
        self.text.crop = crop
        self.text.size = (text_size[0], text_size[1] + 4)

    @property
    def label(self) -> str:
        return self._label

    @label.setter
    def label(self, text: str):
        self._label = text
        self.text.label = self._label


class EditBox(Panel):
    def __init__(
        self,
        value: str,
        width: int = 400,
        height: int = 300,
        left: int = 10,
        top: int = 10,
        align: WindowAlignment = WindowAlignment.FREE,
        theme: Optional[Theme] = None,
        on_change: Optional[Callable] = None,
        multiline: bool = False,
        **kwargs: Any,
    ):
        kwargs["on_click"] = self._focus
        if height < 30:
            height = 30

        super().__init__(
            width=width, height=height, left=left, top=top, align=align, theme=theme, **kwargs,
        )

        self.multiline = multiline
        self.theme.secondary()
        self.on_change = on_change
        self._value = value
        self.text = Text(position=(0, 0, 1), size=(width, height), label=self._value, color=self.theme.text_color,)
        self.add_child("label", self.text)
        self._cursor = -1

    def _on_keypress(self, instr: str):
        new = self.value[0 : self._cursor] + instr + self.value[self._cursor :]
        self._cursor = len(self.value[0 : self._cursor] + instr)
        self.value = new

    def _focus(self):
        self._cursor = len(self.value)
        self._init = False
        # Dummy holder to pass on_click test
        pass

    def cursor_left(self):
        self._cursor -= 1
        if self._cursor < 0:
            self._cursor = 0
        self._init = False

    def cursor_right(self):
        self._cursor += 1
        if self._cursor > len(self.value):
            self._cursor = len(self.value)
        self._init = False

    def backspace(self):
        if self._cursor == 0:
            return
        self.value = self.value[0 : self._cursor - 1] + self.value[self._cursor :]
        self.cursor_left()

    def _exit(self):
        self._cursor = -1
        if callable(self.on_change):
            self.on_change(self._value)
        self._init = False

    def draw(self, **kwargs):
        super().draw()
        if self._cursor > -1:
            label = self._value[0 : self._cursor] + "|" + self._value[self._cursor :]
        else:
            label = self._value
        self.text.label = label
        if self.multiline:
            self.text.wrap(self.size[0])
        text_size = list(self.text.text_size)
        x = 1
        y = (self.size[1] - text_size[1]) / 2
        if y < 0:
            y = 0
        y += 4
        self.text.position = [x, y]
        crop = [0, 0, text_size[0], text_size[1] + 4]
        if text_size[0] > self.size[0]:
            crop[0] = text_size[0] - self.size[0]
            crop[2] = text_size[0]

        if text_size[1] > self.size[1]:
            crop[1] = text_size[1] - self.size[1]
            crop[3] = self.text.text_size[1]

        self.text.crop = crop
        self.text._init = False
        self.text._init_text = False

    @property
    def value(self) -> str:
        return self._value

    @value.setter
    def value(self, text: str):
        self._value = text
        self.text.label = self._value
        if self.text.text_size[0] > self.size[0] and self.multiline:
            self.text.wrap(self.size[0])
        self.refresh()
        self._init = False
